(defproject less-awful-ssl "1.0.8"
  :description "Get an SSLContext without wanting to rip your hair out."
  :url "http://github.com/aphyr/less-awful-ssl"
  :license {:name "Eclipse Public License"
            :url "http://www.eclipse.org/legal/epl-v10.html"}
  :profiles {:dev {:dependencies [[org.clojure/clojure "1.12.0"]]}})
